// Rouslan Dimitrov 2011. This is free software - use it to learn or modify it to suit your needs.
#define _CRT_SECURE_NO_DEPRECATE
#define WIN32_LEAN_AND_MEAN
#pragma warning (disable:4996)
#pragma warning (disable:4244)

#include <windows.h>
#include <vector>
#include <iostream>
#include "midifile.h"
#include "../synth/declarations.h"
#include "events.h"

const u32	MICROSECONDS_PER_MINUTE = 60000000;				// standard MIDI resolution
const u32	TICKS_PER_BEAT = 120;							// standard MIDI timing
const u32	DEFAULT_MPQN = MICROSECONDS_PER_MINUTE / 120;	// 120 BPM; MPQN = micro-seconds per quarter note

using namespace std;

float mpqn2controlsPerTick(u32 mpqn)
{
	u32 BPM = MICROSECONDS_PER_MINUTE / mpqn;
	u32 ticksPerSec = TICKS_PER_BEAT  * BPM / 60;
	return (float)CONTROL_RATE / (float)ticksPerSec;
}

void ConvertTimeToSampleDeltas(EvVec& events)
{
	float controlsPerTick = mpqn2controlsPerTick(DEFAULT_MPQN);
	u32 curPitch = 0;
	u32 prevTime = 0;

	vector<Event*> out;
	out.reserve(events.size());

	for(size_t i=0; i<events.size(); i++) {
		Event *e = events[i];
		if(e->type == E_TEMPO) {
			controlsPerTick = mpqn2controlsPerTick(((EvTempo*)e)->mpqn);
			delete e;
		} else {
			e->time -= prevTime;		// convert to time delta
			prevTime += e->time;
			e->time = controlsPerTick * (float)e->time + 0.4f;	// convert time to samples

			// special processsing
			if(e->type == E_NOTE_ON || e->type == E_NOTE_ON_LAST_VOL) {
				EvNoteOn *curNote = (EvNoteOn*)e;
				if(e->channel == 0)
					curPitch = curNote->pitch;
			}
#if 0
			if(e->type == C_PITCH) {
				EvPitch *p = (EvPitch*)(EvControl*)e;
				p->endPitch = semitone2offset(p->ds + curPitch);
				p->ctrl = max(1, 2*(p->endPitch - semitone2offset(curPitch)) / (p->dt * controlsPerTick));
			}
#endif		
			while(e->time > 256) {		// event time is 8 bit and thus events can be spaced at most 256 samples apart
				out.push_back(new EvContinue());
				i++;
				e->time -= 256;
			}
			out.push_back(e);
		}
	}
	events.swap(out);
}

void RemoveRedundantOffEvents(EvVec& events)
{
	if(events.size() <= 3)
		return;

	vector<Event*> out;
	out.reserve(events.size());
	out.push_back(events[0]);

	for(size_t i=1; i<events.size()-1; i++) {
		Event *e = events[i];
		if(e->type == E_NOTE_OFF) {
			// find neighboring E_NOTE_ON
			Event *e_prev = events[i-1];
			// Make sure off event is for the current note
			if( e_prev->type == E_NOTE_ON) {
				EvNoteOn *on = (EvNoteOn*)e_prev;
				EvNoteOff *off = (EvNoteOff*)e;
				if(on->pitch != off->pitch) {
					delete e;
					e_prev->type = E_NOTE_ON_PORTAM; // notes overlap - portamento
					continue;
				}
			}
			Event *e_next = events[i+1];
			// if there is an note following shortly, we can skip the off event
			if( e_next->type == E_NOTE_ON ) {
				if( e_next->time - e->time <= 14 ) {
					delete e;
					continue;
				}
			}
		}

		out.push_back(events[i]);
	}
	out.push_back(events.back());
	events.swap(out);
}

// It is not legal to have pitch-bend event before the first note
void ComputeSampleOffsets(EvVec& events) {
	u8 lastNotePitch = 0;
	for(size_t i=0; i<events.size(); i++) {
		Event *e = events[i];
		if(e->type == E_NOTE_ON || e->type == E_NOTE_ON_PORTAM) {
			EvNoteOn *on = (EvNoteOn*)e;
			lastNotePitch = on->pitch;
			on->pitch = semitone2offset(lastNotePitch);
		} else if(e->type == C_PITCH) {
			if(!lastNotePitch)
				cout << "Warn: pitch before first note.\n";
			
			EvPitch *p = (EvPitch*)e;
			float fPitch = (float)lastNotePitch + 2.0f/127.0f*(float)p->pitch;
			p->pitch = semitone2offset(fPitch);
		}
	}
}

#if 0
void LinApprox(CtrlVec& out, CtrlVec& events, size_t start, size_t end)
{
	if(start != end)
	{
		EvControl *ev = events[start];

		ev->ds = 2.0f / 256.0f * (events[end]->ctrl - ev->ctrl); // difference in semitones
		ev->dt = (events[end]->time - ev->time);

		out.push_back(ev);

		// end of slide
		//ev = events[end];
		//ev->ds = 0;
		//ev->dt = 1;
		//out.push_back(ev);

		for(size_t i=start+1; i<end+1; i++)
			delete events[i];
	}
}

void LinearApproxEvents(CtrlVec& events)
{
	if(!events.size())
		return;
	vector<EvControl*> out;
	size_t anchor = 0;
	EvControl* max = events[0];

	for(size_t i=1; i<events.size(); i++) {
		EvControl *cur = events[i];
		
		if(cur->ctrl > max->ctrl) {
			max = cur;
		} else {
			LinApprox(out, events, anchor, i-1);
			anchor = i;
			max = cur;
		}
	}

	if(anchor != events.size() - 1)
		LinApprox(out, events, anchor, events.size() - 1);

	events.swap(out);
}
#endif

void CompressNoteVolumeAndPan(EvVec& events)
{
	vector<Event*> out;
	out.reserve(events.size());

	bool pan_changed = false;
	int pan_mask = 0x3;

	for(size_t i=0; i<events.size(); i++) {
		Event *ev = events[i];
		Channel *chan = &g_channels[ev->channel];
		if(ev->type == E_CONTROL) {
			EvMidiControl *ctrl = (EvMidiControl*)ev;
			if(ctrl->ctype == C_PAN) {
				u8 left_on = ctrl->ctrl < 192;
				u8 right_on = ctrl->ctrl > 64;
				u8 new_mask = right_on << 1 | left_on;
				if(pan_mask != new_mask) {
					pan_changed = true;
					pan_mask = new_mask;
				}
				continue; // Also removes the pan from the events list
			}
		} else if(ev->type == E_NOTE_OFF) {
			chan->volume = 0;
		} else if(ev->type == E_NOTE_ON) {
			EvNoteOn *note = (EvNoteOn*)ev;
			if(note->vol != chan->volume || pan_changed) {
				chan->volume = note->vol;
				note->vol = note->vol & 0xFC | pan_mask;
				pan_changed = false;
			} else {
				note->type = E_NOTE_ON_LAST_VOL;
			}
		}
		out.push_back(ev);
	}
	events.swap(out);
}
