// Rouslan Dimitrov 2011. This is free software - use it to learn or modify it to suit your needs.
#include "declarations.h"

#if SIMULATOR
#include <stdio.h>
#else
#include <avr/interrupt.h>
#include <avr/sleep.h>
#endif


// +==========================+
// |    Sound Functionality   |
// +==========================+

extern s8 wave_table1[WAVET_SIZE];
extern s8 wave_table2[WAVET_SIZE];

// Structures

//    +
//   / \-----
//  +
//  IA DH
typedef struct {
	// parameters
	u8	initial;
	u8	attack;
	u8	decay;
	u8	hold;
	u8	init_state;
	// state
	u8	time;
	u8	state;
	u8	value;
} envelope;

typedef struct {
	u16		ds;				// sample delta
	u16		cur_sample;
	envelope blend;			// lerp betweem the wave tables
	u8		mod;
} oscillator;

typedef struct {
	//oscillator osc_m;
	oscillator	osc_c;
	envelope	env_vol;
	u8		vol;
	u8		pan_mask[2];		// left & right enable / disable
	//u8	type;
} channel;

channel g_channels[NUM_CHANNELS];

// Functions

inline void env_reset(envelope *env) {
	env->state = env->init_state;
	env->time = 0;
}

inline void env_update(envelope *env) {
	u8 delta;

	switch(env->state) {
	case ENV_ATTACK:
		delta =  env->time << env->attack;
		env->value = env->initial + delta;
		env->time++;
		if(env->value >= 255 - (2 << env->attack)) {
			env->state = ENV_DECAY;
			env->time = 0;
		}
		break;
	case ENV_DECAY:
		delta = env->time << env->decay;
		env->value = 255 - delta;
		env->time++;
		if(env->value < env->hold) {
			env->state = ENV_HOLD;
			env->value = env->hold;
		}
		break;
	case ENV_HOLD:
		env->value = env->hold;
		break;
	}
}

inline void env_update_all() {	// used only once hence inline
	u8 i;
	for(i=0 ; i<NUM_CHANNELS-1; i++) {
		channel *chan = &g_channels[i];

		env_update(&chan->env_vol);
		chan->env_vol.value = chan->env_vol.value * chan->vol >> 8;

		env_update(&chan->osc_c.blend);
	}
	env_update(&g_channels[3].osc_c.blend);
}

// Functions below are called at SAMPLE_RATE Hz

inline void update_osc(oscillator* osc) {
	osc->cur_sample = osc->cur_sample + osc->ds;
}

inline s8 sample_osc(oscillator* osc) {
	s8 s1, s2;
	u8 sample_pos;
		
	sample_pos = osc->cur_sample >> 9 & (WAVET_SIZE-1);

	s1 = wave_table1[ sample_pos ];
	s2 = wave_table2[ sample_pos ];

	return s1 + ( s2*osc->blend.value >> 8 );
}

inline s8 sample_drum(channel *chan) {
	if(chan->osc_c.ds < 32 || chan->vol == 0)
		return 0;
	chan->osc_c.ds--;

	return sample_osc(&chan->osc_c);
}
// End fast functions

// +==========================+
// |    Song Functionality    |
// +==========================+

extern u8 *song_data;

u16 song_pos;
u16	repeat_pos;
u8	repeats;
u8	time2event;
u16	time2control;
#if SIMULATOR
volatile u8 control_signal = 1;
#else
#define control_signal GPIOR0
#endif

u8	sin_index;
u8	chan0_mod;
const s8 sin_dif_table[16] = {5,14,21,24,24,21,14,5,-5,-14,-21,-24,-24,-21,-14,-5};

// menu variables
u8	volume = 128;
u16	trigger_val = SAMPLE_RATE / CONTROL_RATE;
char *version_string = "R_Synth v2";

u8 read_song_byte(u16 pos);
u16 read_song_word(u16 pos);

inline u8 read_byte() {
#if SIMULATOR
	return song_data[song_pos++];
#else
	return read_song_byte(song_pos++);
#endif
}

inline u16 read_word() {
#if SIMULATOR
	u16 *p = (u16*)(song_data + song_pos);
	song_pos += 2;
	return *p;
#else
	u16 w = read_song_word(song_pos);
	song_pos += 2;
	return w;
#endif
}

// Mark a region with C_REPEAT=0 and C_REPEAT=x to repeat it x times.
inline void handle_repeat(u8 repeats_req) {
	if(!repeats_req) {
		repeat_pos = song_pos;
		repeats = 0;
	} else if(repeats_req > repeats) {
		song_pos = repeat_pos;
		repeats++;
	}		
}

inline void read_envelope(envelope *env) {
	env->init_state = read_byte();
	switch(env->init_state) {
		case ENV_ATTACK:
			env->initial = read_byte();
			env->attack = read_byte();
		case ENV_DECAY:
			env->decay = read_byte();
		case ENV_HOLD:
			env->hold = read_byte();
	}
}

inline void read_events() {
	u8 header = read_byte();
	channel *chan = &g_channels[ header & 0xF ];

	switch(header >> 4) {
	case E_NOTE_ON:
		chan->vol = read_byte();
		chan->pan_mask[0] = (chan->vol & 0x1) ? 0xFF : 0;
		chan->pan_mask[1] = (chan->vol & 0x2) ? 0xFF : 0;
		chan->vol &= 0xFC;
	case E_NOTE_ON_LAST_VOL:
		env_reset(&chan->env_vol);
		env_reset(&chan->osc_c.blend);
	case C_PITCH:
		//cli(); // no need to guard word reads
		chan->osc_c.ds = read_word();
		//sei();
		break;
	case C_MOD:
		chan0_mod = read_byte();
		break;
	case E_NOTE_OFF:
		chan->vol = 0;
		break;
	case C_BLEND: // override blend
		chan->osc_c.blend.state = ENV_FORCED;
		chan->osc_c.blend.hold = read_byte();
		break;
	//case E_PROGRAM:
	//	chan->type = read_byte();
	//	break;
	case C_REPEAT:
		handle_repeat(read_byte());
		break;
	case E_VOL_ENV_SET:
		read_envelope(&chan->env_vol);
		break;
	case E_BL_ENV_SET:
		read_envelope(&chan->osc_c.blend);
		break;
	case E_END:
		song_pos = 0;
		repeats = 0;
		break;
	}
	time2event = read_byte();
}

inline void control() {
#if !SIMULATOR
	while(1) {
		if(control_signal) {
			control_signal = 0;
			PINC = LED_GREEN; // turn on green LED while parsing events
#endif
			oscillator *o;

			while(!time2event) {
				read_events();
			}
			time2event--;

			env_update_all();

			if(chan0_mod) {
				// Channel 0 extras
				o = &g_channels[0].osc_c;
				// Modulation
				o->ds +=  (chan0_mod * sin_dif_table[sin_index++] + 128) >> 8;
				sin_index &= 15;
			}
			menu_update();
			lcd_update();
#if !SIMULATOR
			PORTC &= ~(LED_RED | LED_GREEN);
				
		}
		sleep_cpu();
	}
#endif
}

#if SIMULATOR
void synth_main()
#else
#include <util/delay.h>
__attribute__((naked)) int main(void)
#endif
{
	u8 i;
	for(i = 0; i < NUM_CHANNELS; i++) {
		channel *chan = g_channels + i;
		chan->env_vol.hold = 255;
		chan->env_vol.init_state = ENV_FORCED;
		chan->osc_c.blend.init_state = ENV_FORCED;	
	}
	time2event = read_byte();
	time2control = trigger_val;

	lcd_init(version_string, 0);

#if !SIMULATOR
	// Configure ports
	DDRB = 0b00000110; // OC1AB are L/R PWM output
	PORTB = 0b00001001; // Activate pull-ups for the switches

	DDRC = 0b11011111; // Port C 2 and 3 are LED output, 5 is slider input (ADC)
	PORTC = 0b00;

	DDRD = 0xFE;	// PORTD is LCD out and USART in

	// Set clock to 8 MHz
	//CLKPR = 0x80; // Enable prescaler change
	//CLKPR = 0x80; // Scaling of 1 (20 MHz)

	// Set timer interrupt for sampling thread
	TCCR0A = 0b10; // Clear Timer on Compare (i.e. when TCNT0 == OCR0A)
	TCCR0B = 0b10; // Increment TCNT0 at CPU_SPEED/8 (= 2.5 MHz)
	OCR0A =  CPU_SPEED / 8 / SAMPLE_RATE;
	TIMSK0 = 0b10;

	// Set timer interrupt for control thread
	TCCR1A = 0b10100001; // non-inverting fast PWM on OC1AB
	TCCR1B = 0b00001001; // Increment TCNT2 at CPU_SPEED (20MHz / 256 = 78.125 KHz max sampling rate)
	//OCR1A = 0x80;
	//OCR1B = 0x80;

	// Set sleep mode
	SMCR = 0b0001; // Stop CPU clock (Ch. 10.9)
	PRR = 0b11010100; // Keep only timer 0, 1, ADC and USART on

	ADMUX = 0b01100101; // ADC5, AVcc
	ADCSRA = 0b11000110; // ~32kHz sampling rate

	sei();
	control(); // Never returns
#endif
}

#if SIMULATOR
void generate_sample(u8 *buf) {
#else
ISR(TIMER0_COMPA_vect) {
#endif
	u8 i, smp;
	s8 out[2] = {0,0};
repeat:

	for(i = 0; i < NUM_CHANNELS; i++) {
		update_osc(&g_channels[i].osc_c);
	}


	if(!time2control--) {
		time2control = trigger_val;
		control_signal = 1;
#if SIMULATOR
		control();
#else
		return; // skip a sample (1 in ~250) and yield to the control thread. Gives us +2KHz sampling rate.
#endif
	}

	for(i = 0; i < NUM_CHANNELS - 1; i++) {
		channel *chan = g_channels + i;
		if(chan->env_vol.value) {
			s8 smp = sample_osc(&chan->osc_c) * chan->env_vol.value >> 8;
			out[0] += smp & chan->pan_mask[0];
			out[1] += smp & chan->pan_mask[1];
		}
	}

	smp = sample_drum(g_channels + 3) * g_channels[3].vol >> 8;
	//smp = filter(smp);
	out[0] += smp & g_channels[3].pan_mask[0];
	out[1] += smp & g_channels[3].pan_mask[1];

	out[0] = out[0] * volume >> 8;
	out[1] = out[1] * volume >> 8;

#if SIMULATOR
	buf[0] = out[0] + 128;
	buf[1] = out[1] + 128;
#else
	OCR1A = out[0] + 128;
	OCR1B = out[1] + 128;
	if(TIFR0 & 0b10) { // Save interrupt overhead if timer already triggered Gives us +500Hz sampling rate
		TIFR0 = 0b10;
		PORTC |= LED_RED; // Turn on red LED to show overrun
		goto repeat;
	}
#endif
}
